<div id="{{ figure_id }}">
  <!-- <style scoped> -->
  <!-- {{ default_style }} -->
  <!-- {{ custom_css }} -->
  <!-- </style> -->
  {{ background_svg }}
</div>
<div id="{{ figure_id }}_options">
  <button type="button" class="download btn btn-primary">Save JSON</button>
  <button type="button" class="reactionbutton btn btn-success ">Show Reaction Nodes</button>
  <button type="button" class="svgbutton btn btn-info ">Download SVG</button>
  <!-- <button type="button" class="btn btn-info" href="javascript:(function () { var e = document.createElement('script'); if (window.location.protocol === 'https:') { e.setAttribute('src', 'https://rawgit.com/NYTimes/svg-crowbar/gh-pages/svg-crowbar.js'); } else { e.setAttribute('src', 'http://nytimes.github.com/svg-crowbar/svg-crowbar.js'); } e.setAttribute('class', 'svg-crowbar'); document.body.appendChild(e); })();">Download SVG</button> -->
</div>
<script type="text/Javascript">
{{ javascript_source }}
</script>
